/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Hash File Identification Database Interface
 *
 * This module supports a user identification database saved as a hash file.
 * The user's login name (not including the terminating null) is used as a
 * key to get blocks of data which are terminated by a newline and a null and
 * look like:
 *
 *    <pswd>:<uid>:<gid>:<fullname>:<directory>:<email>:<vc>:<priv>:<epriv>:\n
 *
 * The <pswd> will only be valid if the database is shared with the
 * authentication database.  If the email address has been validated,
 * then <vc> an empty string.  Otherwise, <vc> is the random
 * character code that must be given to validate the email address.
 *
 * For systems that need to be able to look up users by uid number (so far
 * this never happens and it probably never will) this also maintains and
 * uses an alternate database indexed by the user's uid number (as a four
 * byte binary number) and containing lines like:
 *
 *    <pswd>:<login>:<gid>:<fullname>:<directory>:<email>:<vc>:<priv>:<epriv>:\n
 *
 * The <pswd>, <gid>, <uid>, <fullname> and <directory> values should
 * always be identical to the ones in the other hash file.
 */


#include "common.h"
#include "authident.h"

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "set_uid.h"
#include "str.h"
#include "lock.h"
#include "hashfile.h"
#include "entropy.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

char *showopt_ident_module= "ident_hash (db="DBM_NAM_FILE")";


/* MKNAMDBENT - Make a name database entry.  Pass in a BFSZ buffer.  Returns
 * size.
 */

int mknamdbent(char *bf, char *passwd, uid_t uid, gid_t gid,
        char *gecos, char *dir, char *email, char *valcode, int priv, int epriv)
{   
    int rc;
#ifdef SHARE_AUTH_IDENT
    rc= snprintf(bf, BFSZ, "%s:%d:%d:%.300s:%s:%.300s:%s:%d:%d:\n", passwd,
	uid, gid, gecos, dir, email ? email : "", valcode ? valcode : "",
	priv, epriv);
#else
    rc= snprintf(bf, BFSZ, "*:%d:%d:%.300s:%s:%.300s:%s:%d:%d:\n",
	uid, gid, gecos, dir, email ? email : "", valcode ? valcode : "",
	priv, epriv);
#endif
    if (rc < 0 || rc > BFSZ -1)
        die("user fields too long for ident_hash nam db");
}


#ifdef DBM_UID_FILE
/* MKUIDDBENT - Make a name database entry.  Pass in a BFSZ buffer.  Returns
 * size.
 */

int mkuiddbent(char *bf, char *name, char *passwd, gid_t gid,
        char *gecos, char *dir, char *email, char *valcode, int priv, int epriv)
{   
    int rc;
#ifdef SHARE_AUTH_IDENT
    rc= snprintf(bf, BFSZ, "%s:%s:%d:%.300s:%s:%.300s:%s:%d:%d:\n", passwd,
	name, gid, gecos, dir, email ? email : "", valcode ? valcode : "",
	priv, epriv);
#else
    rc= snprintf(bf, BFSZ, "*:%d:%s:%.300s:%s:%.300s:%s:%d:%d:\n",
	name, gid, gecos, dir, email ? email : "", valcode ? valcode : "",
	priv, epriv);
#endif
    if (rc < 0 || rc > BFSZ -1)
        die("user fields too long for ident_hash uid db");
}
#endif /* DBM_UID_FILE */


/* IDENT_GETPWNAM - Get user data from the DBM_NAM_FILE.  Save info into a
 * standard passwd structure in static storage.  Return NULL on failure.
 */

#ifdef HAVEIDENT_GETPWNAM
struct btpasswd *ident_getpwnam(char *nam)
{
    static char bf[BFSZ];
    static char name[MAX_LOGIN_LEN+1];
    static struct btpasswd pw;
    char *e, *b;
    int rc;

    /* Fetch the line from the db file */
    set_httpd_if_shared();
    rc= fetchdbm(DBM_NAM_FILE, nam, strlen(nam), bf, BFSZ);
    set_cfadm_if_shared();

    /* Return NULL if not found */
    if (rc) return NULL;

    /* Copy name into static storage */
    strcpy(name,nam);
    pw.name= name;

    /* Get password */
    if ((e= strchr(bf,':')) == NULL) return NULL;
    pw.passwd= (e-b < 2) ? NULL : bf;
    *e= '\0';

    /* Get uid */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.uid= atoi(b);

    /* Get gid */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.gid= atoi(b);
    
    /* Get fullname */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.gecos= b; *e= '\0';

    pw.email= NULL;
    pw.valcode= NULL;
    pw.priv= -1;
    pw.epriv= -1;
    
    /* Get directory */
    b= e+1; 
    pw.dir= b;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    *e= '\0';

    /* Get email address */
    b= e+1;
    pw.email= b;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    *e= '\0';

    /* Get validation code */
    b= e+1;
    pw.valcode= b;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    *e= '\0';

    /* Get privacy flag */
    b= e+1;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    if (isascii(*b) && isdigit(*b)) pw.priv= atoi(b);

    /* Get email privacy flag */
    b= e+1;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    if (isascii(*b) && isdigit(*b)) pw.epriv= atoi(b);

    return &pw;
}
#endif


/* GETIDENTUID - Get user data from the DBM_UID_FILE.  Save info into a
 * standard passwd structure in static storage.  Return NULL on failure.
 *
 * DBM_UID_FILE entries look like:
 *   <encryptpassswd>:<loginuid>:<gid>:<fullname>:<directory>
 */

#ifdef HAVEIDENT_GETPWUID
struct btpasswd *ident_getpwuid(uid_t uid)
{
    static char bf[BFSZ];
    static struct btpasswd pw;
    char *e, *b;
    int rc;

    /* Fetch the line from the db file */
    set_httpd_if_shared();
    rc= fetchdbm(DBM_UID_FILE, (char *)&uid, sizeof(uid), bf, BFSZ);
    set_cfadm_if_shared();

    /* Return NULL if not found */
    if (rc) return NULL;

    /* Get password */
    if ((e= strchr(bf,':')) == NULL) return NULL;
    pw.passwd= (e-b < 2) ? NULL : bf;
    *e= '\0';

    /* Get name */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.name= b; *e= '\0';

    /* Save uid */
    pw.uid= uid;

    /* Get gid */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.gid= atoi(b);
    
    /* Get fullname */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.gecos= b; *e= '\0';
    
    pw.email= NULL;
    pw.valcode= NULL;
    
    /* Get directory */
    b= e+1; 
    pw.dir= b;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    *e= '\0';

    /* Get email address */
    b= e+1;
    pw.email= b;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    *e= '\0';

    /* Get validation code */
    b= e+1;
    pw.valcode= b;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    *e= '\0';

    /* Get privacy flag */
    b= e+1;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    if (isascii(*b) && isdigit(*b)) pw.priv= atoi(b);

    /* Get email privacy flag */
    b= e+1;
    if (*(e= firstin(b,":\n")) == '\0') return &pw;
    if (isascii(*b) && isdigit(*b)) pw.epriv= atoi(b);

    return &pw;
}
#endif


/* IDENT_WALK - Return the "first" user in the dbm file.
 */

#ifdef HAVEIDENT_WALK
char *ident_walk(int flag)
{
    static char bf[MAX_LOGIN_LEN+1];

    if (!flag)
    {
	set_httpd_if_shared();
	if (walkdbm(DBM_NAM_FILE))
	{
	    set_cfadm_if_shared();
	    die("ident_walk cannot open name db "DBM_NAM_FILE" to read");
	}
	set_cfadm_if_shared();
	
	if (firstdbm(bf, MAX_LOGIN_LEN+1))
	    return NULL;
	else
	    return bf;
    }
    else
    {
	if (nextdbm(bf, MAX_LOGIN_LEN+1))
	    return NULL;
	else
	    return bf;
    }
}
#endif


/* IDENT_SEEK - Set the scan to start after the named user.  Future nextusr()
 * calls will continue from there.
 */

#ifdef HAVEIDENT_SEEK
void ident_seek(char *login)
{
    set_httpd_if_shared();

    if (walkdbm(DBM_NAM_FILE))
    {
	set_cfadm_if_shared();
	die("ident_seek cannot open name db "DBM_NAM_FILE" to read");
    }
    set_cfadm_if_shared();
    
    seekdbm(login, strlen(login));
}
#endif


/* NEWFNAME - change the user's fullname to the given value.
 */

#ifdef HAVEIDENT_NEWFNAME
void ident_newfname(struct btpasswd *pw, char *newfname)
{
    char bf[BFSZ];
    int len;

    set_httpd_if_shared();

    len= mknamdbent(bf, pw->passwd, pw->uid, pw->gid, newfname, pw->dir,
	pw->email, pw->valcode,pw->priv,pw->epriv);

    if (storedbm(DBM_NAM_FILE,
    		 pw->name, strlen(pw->name),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
    }

#ifdef DBM_UID_FILE
    len= mkuiddbent(bf, pw->passwd, pw->name, pw->gid, newfname, pw->dir,
	pw->email, pw->valcode,pw->priv,pw->epriv);

    if (storedbm(DBM_UID_FILE,
    		 (char *)&pw->uid, sizeof(uid_t),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_UID_FILE" (%s)",errdbm());
    }
#endif

    set_cfadm_if_shared();
    flushusercache();
}
#endif /* HAVEIDENT_NEWFNAME */


/* IDENT_NEWGID - change the user's primary gid to the given value.
 */

#ifdef HAVEIDENT_NEWGID
void ident_newgid(struct btpasswd *pw, gid_t newgid)
{
    char bf[BFSZ];
    int len;

    set_httpd_if_shared();

    len= mknamdbent(bf, pw->passwd, pw->uid, newgid, pw->gecos, pw->dir,
        pw->email, pw->valcode,pw->priv,pw->epriv);

    if (storedbm(DBM_NAM_FILE,
    		 pw->name, strlen(pw->name),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
    }

#ifdef DBM_UID_FILE
    len= mkuiddbent(bf, pw->passwd, pw->name, newgid, pw->gecos, pw->dir,
	pw->email, pw->valcode,pw->priv,pw->epriv);

    if (storedbm(DBM_UID_FILE,
    		 (char *)&pw->uid, sizeof(uid_t),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_UID_FILE" (%s)",errdbm());
    }
#endif

    set_cfadm_if_shared();
    flushusercache();
}
#endif /* HAVEIDENT_NEWGID */


/* IDENT_NEWEMAIL - change the user's email address to the given value.
 */

#ifdef HAVEIDENT_NEWEMAIL
void ident_newemail(struct btpasswd *pw, char *newemail, int newvalid)
{
    char bf[BFSZ];
    int len;
    char *newvalcode= NULL;

    if (newemail[0] != '\0' && !newvalid)
        newvalcode= get_noise();

    set_httpd_if_shared();

    len= mknamdbent(bf, pw->passwd, pw->uid, pw->gid, pw->gecos, pw->dir,
        newemail, newvalcode,pw->priv,pw->epriv);

    if (storedbm(DBM_NAM_FILE,
    		 pw->name, strlen(pw->name),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
    }

#ifdef DBM_UID_FILE
    len= mkuiddbent(bf, pw->passwd, pw->name, pw->gid, pw->gecos, pw->dir,
	newemail, newvalcode,pw->priv,pw->epriv);

    if (storedbm(DBM_UID_FILE,
    		 (char *)&pw->uid, sizeof(uid_t),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_UID_FILE" (%s)",errdbm());
    }
#endif

    set_cfadm_if_shared();
    flushusercache();
}
#endif /* HAVEIDENT_NEWGID */


/* IDENT_NEWPRIV - change the user's privacy flags to the given values.
 */

#ifdef HAVEIDENT_NEWPRIV
void ident_newpriv(struct btpasswd *pw, int priv, int epriv)
{
    char bf[BFSZ];
    int len;

    set_httpd_if_shared();

    len= mknamdbent(bf, pw->passwd, pw->uid, pw->gid, pw->gecos, pw->dir,
        pw->email, pw->valcode, priv, epriv);

    if (storedbm(DBM_NAM_FILE,
    		 pw->name, strlen(pw->name),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
    }

#ifdef DBM_UID_FILE
    len= mkuiddbent(bf, pw->passwd, pw->name, pw->gid, pw->gecos, pw->dir,
	pw->email, pw->valcode,priv,epriv);

    if (storedbm(DBM_UID_FILE,
    		 (char *)&pw->uid, sizeof(uid_t),
    		 bf, len+1, 0))
    {
	set_cfadm_if_shared();
    	die("Cannot open user database "DBM_UID_FILE" (%s)",errdbm());
    }
#endif

    set_cfadm_if_shared();
    flushusercache();
}
#endif /* HAVEIDENT_NEWGID */


/* IDENT_ADD - Add an entry to the user database describing a user with
 * the given login, uid, gid, password, fullname, and directory.
 */

#ifdef HAVEIDENT_ADD
void ident_add(char *lid, uid_t uid, gid_t gid, char *fn, char *dir,
        char *email, int valid, int priv, int epriv)
{
    char bf[BFSZ];
    char *valcode= NULL;
    int len;

    if (email != NULL && email[0] != '\0' && !valid)
        valcode= get_noise();

    set_httpd_if_shared();

    len= mknamdbent(bf, NULL, uid, gid, fn, dir, email, valcode, priv, epriv);

    if (storedbm(DBM_NAM_FILE,
		 lid, strlen(lid),	/* don't include terminating NUL */
		 bf, len+1,		/* include terminating NUL */
		 1))			/* create file if necessary */
    {
	set_cfadm_if_shared();
	die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
    }

#ifdef DBM_UID_FILE

    len= mkuiddbent(bf, NULL, lid, gid, fn, dir, email, valcode, priv, epriv);

    if (storedbm(DBM_UID_FILE,
		 (char *)&uid, sizeof(uid_t),
		 bf, len+1,		/* include terminating NUL */
		 1))			/* create file if necessary */
    {
	set_cfadm_if_shared();
	die("Cannot open user database "DBM_UID_FILE" (%s)",errdbm());
    }
#endif

    set_cfadm_if_shared();
}
#endif


/* IDENT_DEL - Delete the named user from the ident database.
 */

#ifdef HAVEIDENT_DEL
void ident_del(char *login)
{
    deletedbm(DBM_NAM_FILE, login, strlen(login), 0);
#ifdef DBM_UID_FILE
    deletedbm(DBM_UID_FILE, (char *)&pw->uid, sizeof(pw->uid), 0);
#endif /* DBM_UID_FILE */
}
#endif
